#include <Python.h>
#include "numpy/arrayobject.h"
#include "structmember.h"
#include <stdio.h>

#include "input.h"
#include "ff.h"

typedef struct {
  PyObject_HEAD
  // object data
  FORCE_FIELD *ff;
} ForceField;

static void
ForceField_dealloc(ForceField *self)
{
  self->ob_type->tp_free((PyObject*)self);
  if (self->ff != NULL) {
    delete_force_field(self->ff);
  }
}

static PyObject *
ForceField_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  ForceField *self;

  self = (ForceField *)type->tp_alloc(type, 0);
  
  return (PyObject *)self;
}

static int
ForceField_init_from_file(ForceField *self, PyObject *args, PyObject *kwds)
{
  char *ff_file_name;
  ATOM *atoms;
  int n, nat;
  CRYSTAL crystal;

  self->ff = NULL;
  
  if (!PyArg_ParseTuple(args, "s", &ff_file_name)) {
    return -1;
  }
  // read force field from file
  printf("reading force field definition from file %s\n", ff_file_name);
  FILE *fh = fopen(ff_file_name, "r");
  if (fh == NULL) {
    PyErr_SetString(PyExc_IOError, "File not found!");
    return -1;
  }
  atoms = read_force_field(fh, &nat, &crystal);
  fclose(fh);
  if (atoms == NULL) {
    PyErr_SetString(PyExc_RuntimeError, "Syntax error in force field file!");
    return -1;
  }
  assign_parameters(atoms, nat, 1);
  
  //
  printf("Atomc in Unit Cell:\n");
  for (n=0; n<nat; n++) {
    printf("%d  %+f %+f %+f\n", atoms[n].atom_type, atoms[n].x, atoms[n].y, atoms[n].z);
  }
  printf("Lattice vectors:\n");
  printf("a  %f %f %f\n", crystal.ax, crystal.ay, crystal.az);
  printf("b  %f %f %f\n", crystal.bx, crystal.by, crystal.bz);
  printf("c  %f %f %f\n", crystal.cx, crystal.cy, crystal.cz);
  //
  
  self->ff = build_force_field(nat, atoms, &crystal, 1); 

  return 0;
}

static int
ForceField_init_from_lists(ForceField *self, PyObject *args, PyObject *kwds)
{
  PyObject *py_atomlist;
  PyObject *py_typelist;
  PyObject *py_lattice_vectors;
  PyObject *py_atomtuple, *py_poslist, *py_posseq;
  int verbose = 0;
  
  ATOM *atoms;
  ATOM *atom;
  int n, nat;
  CRYSTAL crystal;

  static char *kwlist[] = {"atomlist", "atomtypes", "lattice_vectors", "verbose", NULL};
  
  self->ff = NULL;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!O!O!|i", kwlist,
			&PyList_Type, &py_atomlist, 
			&PyList_Type, &py_typelist, 
			&PyList_Type, &py_lattice_vectors,
			&verbose	   )) {
    return -1;
  }
  nat = (int) PyList_Size(py_atomlist);
  if (nat != PyList_Size(py_typelist)) {
    PyErr_SetString(PyExc_ValueError, "There should be one atomtype for each atom!");
    return -1;
  }
  if (3 != PyList_Size(py_lattice_vectors)) {
    PyErr_SetString(PyExc_ValueError, "3 lattice vectors are required!");
    return -1;
  }
  // extract all atoms
  atoms = (ATOM *) malloc(sizeof(ATOM)*nat);
  for(n=0;n<nat;n++) {
    py_atomtuple = PyList_GetItem(py_atomlist, (Py_ssize_t) n);
    py_poslist = PyTuple_GetItem(py_atomtuple, (Py_ssize_t) 1);
    py_posseq = PySequence_Fast(py_poslist, "Atom positions should be a list [x,y,z] or tuple (x,y,z)!");

    atom = &(atoms[n]);

    atom->x = PyFloat_AsDouble(PySequence_GetItem(py_posseq, (Py_ssize_t) 0));
    atom->y = PyFloat_AsDouble(PySequence_GetItem(py_posseq, (Py_ssize_t) 1));
    atom->z = PyFloat_AsDouble(PySequence_GetItem(py_posseq, (Py_ssize_t) 2));
    
    Py_DECREF(py_posseq);
    //
    
    atom->atom_type = (int) PyInt_AsLong(PyList_GetItem(py_typelist, (Py_ssize_t) n));
    if ((atom->atom_type < 0) || (atom->atom_type >= dreiding_nr_atom_types)) {
      PyErr_SetString(PyExc_ValueError, "Invalid atom type!");
      free(atoms);
      return -1;
    }
  }
  // read lattice vectors
  py_poslist = PyList_GetItem(py_lattice_vectors,(Py_ssize_t) 0);
  crystal.ax = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 0));
  crystal.ay = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 1));
  crystal.az = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 2));
  py_poslist = PyList_GetItem(py_lattice_vectors,(Py_ssize_t) 1);
  crystal.bx = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 0));
  crystal.by = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 1));
  crystal.bz = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 2));
  py_poslist = PyList_GetItem(py_lattice_vectors,(Py_ssize_t) 2);
  crystal.cx = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 0));
  crystal.cy = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 1));
  crystal.cz = PyFloat_AsDouble(PyList_GetItem(py_poslist, (Py_ssize_t) 2));
  // assign atom types
  assign_parameters(atoms, nat, verbose);
  
  //
  if (verbose > 0) {
    printf("Atomc in Unit Cell:\n");
    for (n=0; n<nat; n++) {
      printf("%d  %+f %+f %+f\n", atoms[n].atom_type, atoms[n].x, atoms[n].y, atoms[n].z);
    }
    printf("Lattice vectors:\n");
    printf("a  %f %f %f\n", crystal.ax, crystal.ay, crystal.az);
    printf("b  %f %f %f\n", crystal.bx, crystal.by, crystal.bz);
    printf("c  %f %f %f\n", crystal.cx, crystal.cy, crystal.cz);
  }
  // create lists of bonds, angles, dihedrals, improper torsions and vdW interactions
  self->ff = build_force_field(nat, atoms, &crystal, verbose); 

  return 0;
}

static PyObject *
ForceField_getEnergyAndGradient(ForceField *self, PyObject *args)
{
  PyArrayObject *coords_arr, *grads_arr;
  int dimensions[1];
  int ndim, n, nat;
  double *x_ptr, *y_ptr, *z_ptr;
  double x,y,z;
  double energy;
  
  if (!PyArg_ParseTuple(args, "O!", &PyArray_Type, &coords_arr)) {
    return NULL;
  }

  if (coords_arr->nd != 1 || coords_arr->descr->type_num != PyArray_DOUBLE) {
    PyErr_SetString(PyExc_ValueError, "Coordinates vector must be one-dimensional and of type double!");
    return NULL;
  }
  ndim = coords_arr->dimensions[0];
  nat = get_number_of_atoms(self->ff);
  if (ndim != 3*nat) {
    PyErr_SetString(PyExc_ValueError, "Wrong number of coordinates!");
    return NULL;
  }
  // update coordinates of atoms in the force field
  for(n=0; n<nat; n++) {
    x = *(double *) (coords_arr->data + (3*n+0)*coords_arr->strides[0]);
    y = *(double *) (coords_arr->data + (3*n+1)*coords_arr->strides[0]);
    z = *(double *) (coords_arr->data + (3*n+2)*coords_arr->strides[0]);
    // set coordinates of n-th atom
    set_force_field_coordinates(self->ff, n, x,y,z);
  }

  evaluate_force_field(self->ff);
  energy = collect_force_field(self->ff);

  // copy gradient back
  dimensions[0] = ndim;
  grads_arr = (PyArrayObject *) PyArray_FromDims(1, dimensions, PyArray_DOUBLE);
  if (grads_arr == NULL) {
    return NULL;
  }

  for(n=0; n<nat; n++) {
    x_ptr = (double *) (grads_arr->data + (3*n+0)*grads_arr->strides[0]);
    y_ptr = (double *) (grads_arr->data + (3*n+1)*grads_arr->strides[0]);
    z_ptr = (double *) (grads_arr->data + (3*n+2)*grads_arr->strides[0]);

    // get gradient on n-th atom
    get_force_field_gradient(self->ff, n, &x, &y, &z);
    *x_ptr = x;
    *y_ptr = y;
    *z_ptr = z;
  }

  return Py_BuildValue("(fO)", (double) energy, (PyObject *) grads_arr);
}

static PyMethodDef ForceField_methods[] = {
  {"getEnergyAndGradient", (PyCFunction)ForceField_getEnergyAndGradient, METH_VARARGS,
"Computes the energy and gradient on the atoms in the central unit cell.\n\
The number and order of atoms should remain the same as used when defining the force field.\n\n\
Parameters:\n\
===========\n\
coords: numpy array with shape (3*Nat), coords[3*i:3*(i+1)] are the x,y,z coordinates of the i-th atom in bohr\n\
\n\
Returns:\n\
========\n\
energy: total force field energy in Hartree\n\
grads: numpy array with shape (3*Nat), grads[3*i:3*(i+1)] is the gradient on the i-th atom in Hartree/bohr\n\
    "},
  {NULL, NULL, 0, NULL}  /* Sentinel */
};

static PyTypeObject ForceFieldType = {
  PyObject_HEAD_INIT(NULL)
  0,                         /*ob_size*/
  "ff.ForceField",             /*tp_name*/
  sizeof(ForceField),             /*tp_basicsize*/
  0,                         /*tp_itemsize*/
  (destructor)ForceField_dealloc, /*tp_dealloc*/
  0,                         /*tp_print*/
  0,                         /*tp_getattr*/
  0,                         /*tp_setattr*/
  0,                         /*tp_compare*/
  0,                         /*tp_repr*/
  0,                         /*tp_as_number*/
  0,                         /*tp_as_sequence*/
  0,                         /*tp_as_mapping*/
  0,                         /*tp_hash */
  0,                         /*tp_call*/
  0,                         /*tp_str*/
  0,                         /*tp_getattro*/
  0,                         /*tp_setattro*/
  0,                         /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Builds a forces field with periodic boundary conditions.\n\
valence terms:\n\
  bond stretching, angle bending, torsion and spectroscopic inversion\n\
non-bonding terms:\n\
  van der Waals (Lennard Jones potential)\n\
The Coulomb interaction is not included, so all atoms should be approximately neutral.\n\
The parameters are taken from the DREIDING force field. \n\
\n\
So far only two atom-types exist:\n\
   0: hydrogen\n\
   1: aromatic carbon\n\
\n\
Parameters:\n\
===========\n\
atomlist: geometry which defined the bonding topology, \n\
  list of tuples (Zat,[x,y,z]) for each atom           \n\
atomtypes: list with indeces of atom types             \n\
lattice_vectors: 3 lists with translation vectors of lattice\n\
  [[ax,ay,az],[bx,by,bz],[cx,cy,cz]]\n\
verbose: print information about the\n\
",      /* tp_doc */
  0,                         /* tp_traverse */
  0,                         /* tp_clear */
  0,               /* tp_richcompare */
  0,               /* tp_weaklistoffset */
  0,               /* tp_iter */
  0,               /* tp_iternext */
  ForceField_methods,             /* tp_methods */
  0,             /* tp_members */
  0,                         /* tp_getset */
  0,                         /* tp_base */
  0,                         /* tp_dict */
  0,                         /* tp_descr_get */
  0,                         /* tp_descr_set */
  0,                         /* tp_dictoffset */
  (initproc)ForceField_init_from_lists,      /* tp_init */
  0,                         /* tp_alloc */
  ForceField_new,                 /* tp_new */
};

static PyMethodDef module_methods[] = {
  {NULL, NULL, 0, NULL}  /* Sentinel */
};

#ifndef PyMODINIT_FUNC/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
initff(void)
{
  PyObject* m;

  if (PyType_Ready(&ForceFieldType) < 0)
    return;

  m = Py_InitModule3("ff", module_methods,
		     "Force Field for periodic QM/MM calculations of molecular crystals.");
  import_array();
  if (m == NULL)
    return;

  Py_INCREF(&ForceFieldType);
  PyModule_AddObject(m, "ForceField", (PyObject *)&ForceFieldType);
}

