"""
numerically integrate eqn. (3) from
Phys. Status Solidi B 249, No 2 (2012)
"""
import numpy as np
import scipy
from scipy import integrate

def gamma_lr_exact(x, Rlr, tA, tB):
    # integrand
    def integrand(q):
        f = 2*(tA*tB)**4/(x*np.pi) * np.sin(q*x)/(q*(q**2+tA**2)**2*(q**2+tB**2)**2) * np.exp(-q**2*Rlr**2/4.0)
        print f
        return f
    I = integrate.quad(integrand, 0.0, scipy.Inf)
    return I

def gamma_lr_exact_arr(x, Rlr, tA, tB):
    I = [gamma_lr_exact(xi, Rlr, tA, tB) for xi in x]
    return np.array(I)

def gamma_lr_approx(x, Rlr):
    return scipy.special.erf(x/Rlr)*scipy.special.erf(x)/x

class gamma_lr_integral:
    def __init__(self, Rlr):
        self.Rlr = Rlr
        self.x = np.linspace(0.000001, 20.0, 200)
        self.g = gamma_lr_exact_arr(self.x, self.Rlr)
        # interpolate

########## SWITCHING FUNCTIONS #############

def f_erf(w, r):
    f = scipy.special.erf(w*r)
    return f

def f_erfgau(w, r):
    f = scipy.special.erf(w*r) - 2.0/np.sqrt(np.pi) * (w*r) * np.exp(-(w*r)**2 / 3.0)
    return f

def Fourier_erf(w,q):
    ft = 2.0j/q * np.exp(-q**2 / (4.0*w**2))
    return ft

def Fourier_erfgau(w,q):
    ft = 2.0j/q * np.exp(-q**2 / (4.0*w**2)) \
        - pow(3,1.5)*1.0j/w**2 * q * np.exp(-3.0/4.0 * q**2 / w**2)
    return ft

def gamma_lr_integral_Fslater(R, w, ft, tA, tB):
    def integrand(q):
        intg = (tA*tB)**4 / (np.pi*R) * 1.0/1.0j * np.sin(q*R) / ((q**2 + tA**2)**2 * (q**2 + tB**2)**2) * ft(w, q)
        return intg
    I = integrate.quad(integrand, 0.0, scipy.Inf)
    return I[0]

def gamma_lr_integral_Fslater_arr(R, w, ft, tA, tB):
    I = [gamma_lr_integral_Fslater(Ri, w, ft, tA, tB) for Ri in R]
    return np.array(I)

def gamma_lr_integral_Fgauss(R, w, ft, sA, sB):
    def integrand(q):
        intg = 1.0 / (np.pi*R) * 1.0/1.0j * np.exp(-0.5*(sA**2 + sB**2)*q**2) * np.sin(q*R) * ft(w, q)
        return intg
    I = integrate.quad(integrand, 0.0, scipy.Inf)
    return I[0]

def gamma_lr_integral_Fgauss_arr(R, w, ft, sA, sB):
    I = [gamma_lr_integral_Fgauss(Ri, w, ft, sA, sB) for Ri in R]
    return np.array(I)

    
def gamma_lr_approx(R, w, f, tA, tB):
    cAB = 5.0/16.0 * np.sqrt(np.pi/2.0) * tA
    glr = f(w,R) * scipy.special.erf(cAB*R)/R
    return glr

if __name__ == "__main__":
    from matplotlib import pyplot as plt
    tA = 2.0
    tB = 2.0
    
    ############### SWITCHING FUNCTIONS #############
    R = np.linspace(0.5, 20.0, 200)
    # error function
    plt.cla()
    plt.ylim((0.0, 0.65))
    plt.xlabel("$R_{AB}$ / bohr", fontsize=17)
    Coulomb = 1.0/R
    w = 0.3
    plt.plot(R, Coulomb, label="1/r", lw=2)
    plt.plot(R, f_erf(w,R)/R, label="long range", lw=2)
    plt.plot(R, (1.0-f_erf(w,R))/R, label="short range", lw=2)
#    plt.text(11.0, 0.3, "f(r) = erf($\omega r$)", fontsize=20)
    plt.legend()
    plt.savefig("erf_switching.png")
    plt.show()

    # erfgau
    plt.cla()
    plt.ylim((0.0, 0.65))
    plt.xlabel("$R_{AB}$ / bohr", fontsize=17)
    Coulomb = 1.0/R
    w = 0.3
    plt.plot(R, Coulomb, label="1/r", lw=2)
    plt.plot(R, f_erfgau(w,R)/R, label="long range", lw=2)
    plt.plot(R, (1.0-f_erfgau(w,R))/R, label="short range", lw=2)
#    plt.text(11.0, 0.3, "f(r) = erfgau($\omega r$)", fontsize=20)
    plt.legend()
    plt.savefig("erfgau_switching.png")
    plt.show()


    ############### GAMMA LONG-RANGE ################
    R = np.linspace(0.000001, 20.0, 200)
    # error function
    plt.cla()
    plt.xlabel("$R_{AB}$ / bohr", fontsize=17)
    plt.ylabel("$\gamma^{lr}$ / Hartree", fontsize=17)
    colours = ["red", "blue", "green"]
    for iw, w in enumerate([0.1, 0.2, 0.33]):
        glr_exact = gamma_lr_integral_Fslater_arr(R, w, Fourier_erf, tA, tB).real
        glr_approx = gamma_lr_approx(R, w, f_erf, tA, tB)
        
        plt.plot(R,glr_exact, color=colours[iw], lw=3, label="$\omega = %s$" % w)
        plt.plot(R,glr_approx, color=colours[iw], lw=4, ls="-.")
    plt.legend()
    plt.savefig("gamma_lr_switch_erf.png")
    plt.show()

    plt.clf()
    # error function with gaussian
    plt.xlabel("$R_{AB}$ / bohr", fontsize=17)
    plt.ylabel("$\gamma^{lr}$ / Hartree", fontsize=17)
    colours = ["red", "blue", "green"]
    for iw, w in enumerate([0.1, 0.2, 0.3]):
        glr_exact_erf = gamma_lr_integral_Fslater_arr(R, w, Fourier_erf, tA, tB).real
        glr_exact = gamma_lr_integral_Fslater_arr(R, w, Fourier_erfgau, tA, tB).real
        glr_approx = gamma_lr_approx(R, w, f_erfgau, tA, tB)
        
        plt.plot(R,glr_exact_erf, color=colours[iw], lw=1, ls=":")
        plt.plot(R,glr_exact, color=colours[iw], lw=3, label="$\omega = %s$" % w)
        plt.plot(R,glr_approx, color=colours[iw], lw=4, ls="-.")
    plt.legend()
    plt.savefig("gamma_lr_switch_erfgau.png")
    plt.show()
    

