#!/usr/bin/env python
from DFTB import AtomicData
from DFTB import XYZ
import numpy as np
from numpy.linalg import norm
import itertools

from DFTB.Bandstructure import route_through_zone, plot_1d_bandstructure, write_2d_bandstructure
from DFTB import Periodic

a0_graphene = 1.42

def symmetry_points():
    a = a0_graphene / AtomicData.bohr_to_angs

    Gamma = np.array([0.0, 0.0, 0.0])
    K1 = np.array([0.0, -4.0*np.pi/(3.0*np.sqrt(3.0)*a), 0.0])
    K2 = np.array([2.0*np.pi/(3.0*a), -2.0*np.pi/(3.0*np.sqrt(3.0)*a), 0.0])
    K3 = np.array([2.0*np.pi/(3.0*a), +2.0*np.pi/(3.0*np.sqrt(3.0)*a), 0.0])

    M = np.array([2.0*np.pi/(3.0*a), 0.0, 0.0])
#    return [K3, Gamma, M, K3] #,K1, K2, K3, Gamma]
    return [Gamma, K3] #,K1, K2, K3, Gamma]

def graphene_lattice():
    # see http://oer.physics.manchester.ac.uk/AQM2/Notes/Notes-6.4.html
    a = a0_graphene / AtomicData.bohr_to_angs
    # lattice vectors
    a1 = a / 2.0 * np.array([3.0, +np.sqrt(3.0), 0.0])
    a2 = a / 2.0 * np.array([3.0, -np.sqrt(3.0), 0.0])
    a3 = 100000.0 * np.array([0.0, 0.0, 1.0])
    lattice_vectors = [a1,a2, a3]
    # basis atoms
    d1 = a / 2.0 * np.array([1.0, np.sqrt(3.0), 0.0])
    unitcell = [(6, np.array([0.0,0.0,0.0])), (6, d1)]
    # brillouin zone
    lat = Periodic.CrystalSystem()
    lat.setLatticeVectors([a1,a2,a3])
    b1,b2,b3 = lat.getLatticeVectors(space="reciprocal")
    return lat, unitcell

def graphene():
    lat, unitcell = graphene_lattice()
    a1,a2,a3 = lat.getLatticeVectors(space="real")
    lattice_vectors = [a1,a2,a3]
    b1,b2,b3 = lat.getLatticeVectors(space="reciprocal")
    kpoints = []
    kcoordinates = []
    for r in np.linspace(0.0, 1.0, 40):
        for s in np.linspace(0.0, 1.0, 40):
            k = r*b1 + s*b2
            kpoints.append(k)
            kcoordinates.append(np.array([r,s,0.0]))
    return unitcell, lattice_vectors, kpoints, kcoordinates

def graphene_special_points():
    lat, unitcell = graphene_lattice()
    b1,b2,b3 = lat.getLatticeVectors(space="reciprocal")
    Gamma = 0.0*b1
    Sigma = 0.25 * b1
    M = 0.5*b1
    K = (2.0*b1 + b2)/3.0
    Lambda = 0.5 * K
    route = [Gamma, Sigma, M, K, Lambda, Gamma]
    names = ["$\Gamma$", "$\Sigma$", "M", "K", "$\Lambda$", "$\Gamma$"]
    return route, names
    
def carbon_fcc():
    a = 1.42 / AtomicData.bohr_to_angs
    # lattice vectors
    a1 = a * np.array([1,0,0])
    a2 = a * np.array([0,1,0])
    a3 = a * np.array([0,0,1])
    # 
    atomlist = [(6, np.array([0,0,0])),\
                    (6, np.array([0.5,0.5,0])), \
                    (6, np.array([0.5,0,0.5])), \
                    (6, np.array([0,0.5,0.5]))]
    ks = 0

def lattice1D():
    a = 3 * 1.4 / AtomicData.bohr_to_angs
    a1 = a * np.array([1.0, 0.0, 0.0])
    lattice_vectors = [a1]
    # basis atoms
    atomlist = [(6, np.array([0,0,0])), (6, np.array([a/3,0,0])), (6, np.array([a*2.0/3.0,0,0]))]
    # brillouin zone
    kx,ky,kz = np.mgrid[-np.pi/a:np.pi/a:1j*100, 0:0:1j*1, 0:0:1j*1]
    ks = np.vstack((kx.ravel(), ky.ravel(), kz.ravel())).transpose()
    return atomlist, lattice_vectors, ks

if __name__ == "__main__":
    from DFTB.DFTB2 import DFTB2
    from DFTB import utils
    from DFTB.XYZ import read_xyz, extract_keywords_xyz
    from DFTB.Molden import MoldenExporter
    from Analyse.Cube import CubeExporter
    import sys
    """
    # 1D ROUTE through brillouin zone
    atomlist, lattice_vectors, kpoints, kcoordinates = graphene() #lattice1D() #graphene()

#    dftb2 = DFTB2(atomlist, parameter_set="hotbit", verbose=1)
    dftb2 = DFTB2(atomlist, parameter_set="homegrown", verbose=1)
#    dftb2 = DFTB2(atomlist, parameter_set="3ob", verbose=1)
    dftb2.setGeometry(atomlist)

    points, point_names = graphene_special_points()
    seglengths, segboundaries, kpoints = route_through_zone(points)
#    lattice, band_structure = dftb2.runPeriodicNonSCC(lattice_vectors, kpoints, nmax=5)
    lattice, band_structure, dq = dftb2.runPeriodicSCC(lattice_vectors, kpoints, nmax=5)


##    print band_structure
    HOMO, LUMO = dftb2.Nelec_val/2-1, dftb2.Nelec_val/2
    plot_1d_bandstructure(points, point_names, seglengths, segboundaries, band_structure, highlight_bands=[HOMO,LUMO])
    """
    # 2D BAND STRUCTURE
    atomlist, lattice_vectors, kpoints, kcoordinates = graphene() #lattice1D() #graphene()

    dftb2 = DFTB2(atomlist, parameter_set="homegrown", verbose=1)
    dftb2.setGeometry(atomlist)

    dftb2.setSpecialKPoint(kpoints[0])
    lattice, band_structure, dq = dftb2.runPeriodicSCC(lattice_vectors, kpoints, nmax=5)
    XYZ.write_xyz("graphene_lattice.xyz", [lattice])
    write_2d_bandstructure("graphene_band_structure.dat", kcoordinates, band_structure)

    molden = MoldenExporter(dftb2)
    molden.export(molden_file="/tmp/graphene_periodic_Gamma.input")


